@@ -22,9 +22,33 @@
 
 import dataiter
 import itertools
+import math
 import numpy as np
 import string
 
+from dataiter import deco
+
+
+def count_digits(value):
+    if np.isnan(value): return 0, 0
+    if math.isinf(value): return 0, 0
+    parts = str(float(value)).split(".")
+    n = len(parts[0].lstrip("0"))
+    m = len(parts[1].rstrip("0"))
+    return n, m
+
+def format_floats(seq):
+    precision = dataiter.PRINT_FLOAT_PRECISION
+    if any(0 < x < 1/10**precision or x > 10**16 - 1 for x in seq):
+        # Format tiny and huge numbers in scientific notation.
+        f = np.format_float_scientific
+        return [f(x, precision=precision, trim="-") for x in seq]
+    # Format like largest by significant digits.
+    digits = [count_digits(x) for x in seq]
+    n = max(x[0] for x in digits)
+    m = max(x[1] for x in digits)
+    precision = min(m, max(0, precision - n))
+    return [f"{{:.{precision}f}}".format(x) for x in seq]
 
 def generate_colnames(n):
     return list(itertools.islice(yield_colnames(), n))
@@ -32,20 +56,16 @@ def generate_colnames(n):
 def length(value):
     return 1 if np.isscalar(value) else len(value)
 
-def np_to_string(value, quote=True):
-    str_ = format_quoted if quote else str
-    return np.array2string(
-        np.array(value),
-        max_line_width=dataiter.PRINT_MAX_WIDTH,
-        precision=dataiter.PRINT_FLOAT_PRECISION,
-        formatter={
-            "datetime": str_,
-            "numpystr": str_,
-            "str": str_,
-        },
-    )
+@deco.listify
+def pad(strings, align="right"):
+    width = max(len(x) for x in strings)
+    for value in strings:
+        padding = " " * (width - len(value))
+        yield (padding + value
+               if align == "right"
+               else value + padding)
 
-def format_quoted(value):
+def quote(value):
     return '"{}"'.format(str(value).replace('"', r'\"'))
 
 def unique_keys(keys):
